/***********************************************************
        OpenGL P175B512 darbas - Demke

         2008 A-E Software 
        Arvydas Burdulis
        http://ae.gamedev.lt  aab@auto.lt
        
        Licenzija:
                Su kodu galima daryti viska, isskyrus antrasciu modifikavima.
        

        Reikalingi lib'ai:
                kernel32
                user32
                winmm
                opengl32
                glu32
                gdi32
                muziks  (MIDI grojimas)
        
        g++ parametrai exe failo dydzio optimizavimui:
                GCC kompiliatoriaus bendri:
                        -fsingle-precision-constant
                        -fshort_double
                        -fmerge_constants
                C++ kompiliatoriaus:
                        -Os
                Linkerio:
                        -s

        Pradeta:
            2008-09-09
        Paskutini karta redaguota:
            2008-12-11
************************************************************/


//#define DEBUG         //Debug versija
//#define WAV           //Generuoti WAV garsa

#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <math.h>

#ifdef DEBUG
#include <iostream>
#include <iomanip>
using namespace std;
#include "kodas\zinutes.hpp"
#endif

#ifndef M_PI
#define M_PI    3.14159265358979f
#endif
#ifndef M_PI_2
#define M_PI_2  M_PI / 2.0
#endif

HDC                     hDC=NULL;                  //GDI Device Context
HGLRC           hRC=NULL;                  //Permanent Rendering Context
HWND            hWnd=NULL;                 //Window Handle
HINSTANCE       hInstance;                 //Instance Of The Application

#include "muziks.h"
#include "kodas\ogl_ext_constants.hpp"
#include "kodas\ogl_ext.hpp"
#include "kodas\taskas.hpp"
#include "kodas\kintamieji.hpp"
#include "kodas\tekstura.hpp"
#include "kodas\kamera.hpp"
#include "kodas\geom_obj.hpp"
#include "kodas\tekstas.hpp"
#include "kodas\obj\objektyvas.hpp"
#include "kodas\obj\ziedas.hpp"
#include "kodas\obj\kubas_wire.hpp"
#include "kodas\obj\kubas_tekstas.hpp"
#include "kodas\obj\fonas.hpp"
#include "kodas\obj\fonas_kubai.hpp"
#include "kodas\obj\ziedas2.hpp"
#include "kodas\obj\tekstas_demke.hpp"
#include "kodas\laikas.hpp"
#include "kodas\funkcijos.hpp"
#include "midi_data.h"

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);   //WndProc


void LangasNaikinti(GLvoid)                                                             //Naikinti langa
{
        ChangeDisplaySettings(NULL, 0);
        ShowCursor(TRUE);

        if (hRC)                                                                                        //Yra Rendering Context?
        {
                if (!wglMakeCurrent(NULL, NULL))                                //Ar galima atlaisvinti DC ir RC Contextus?
                {
                        #ifdef DEBUG
                        Klaida("Klaida atlaisvinant DC ir RC");
                        #endif
                }

                if (!wglDeleteContext(hRC))                                             //Ar galima naikinti RC?
                {
                        #ifdef DEBUG
                        Klaida("Klaida atlaisvinant Rendering Context");
                        #endif
                }
                hRC = NULL;
        }

        if (hDC && !ReleaseDC(hWnd, hDC))                                       //Ar galima atlaisvinti DC
        {
                #ifdef DEBUG
                Klaida("Klaida atlaisvinant Device Context");
                #endif
                hDC = NULL;
        }

        if (hWnd && !DestroyWindow(hWnd))                                       //Ar galima naikinti langa?
        {
                #ifdef DEBUG
                Klaida("Klaida naikinant langa");
                #endif
                hWnd = NULL;
        }

        if (!UnregisterClass("OpenGL",hInstance))
        {
                #ifdef DEBUG
                Klaida("Klaida atlaisvinant hInstance");
                #endif
                hInstance = NULL;
        }
}
 
bool LangasKurti()
{
        GLuint          PixelFormat;                    //Holds The Results After Searching For A Match
        WNDCLASS        wc;                                             //Windows Class Structure
        DWORD           dwExStyle;                              //Window Extended Style
        DWORD           dwStyle;                                //Window Style
        RECT            WindowRect;                             //Grabs Rectangle Upper Left / Lower Right Values
        WindowRect.left = (long)0;                      //Left
        WindowRect.right = (long)langasX;       //Right
        WindowRect.top = (long)0;                       //Top
        WindowRect.bottom = (long)langasY;      //Bottom

        hInstance                       = GetModuleHandle(NULL);                                //Grab An Instance For Our Window
        wc.style                        = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;   //Redraw On Size, And Own DC For Window.
        wc.lpfnWndProc          = (WNDPROC)WndProc;                                             //WndProc Handles Messages
        wc.cbClsExtra           = 0;                                                                    //No Extra Window Data
        wc.cbWndExtra           = 0;                                                                    //No Extra Window Data
        wc.hInstance            = hInstance;                                                    //Set The Instance
        wc.hIcon                        = LoadIcon(NULL, IDI_WINLOGO);                  //Load The Default Icon
        wc.hCursor                      = LoadCursor(NULL, IDC_ARROW);                  //Load The Arrow Pointer
        wc.hbrBackground        = NULL;                                                                 //No Background Required For GL
        wc.lpszMenuName         = NULL;                                                                 //We Don't Want A Menu
        wc.lpszClassName        = "OpenGL";                                                             //Set The Class Name

        if (!RegisterClass(&wc))                                                                        //Attempt To Register The Window Class
        {
                #ifdef DEBUG
                Klaida("Klaida registruojant lango klase");
                #endif
                
                return false;
        }
        
        //Gauti monitoriaus raiska
        langasX = GetSystemMetrics(SM_CXSCREEN);
        langasY = GetSystemMetrics(SM_CYSCREEN);

        dwExStyle = WS_EX_APPWINDOW;                                                    //Window Extended Style
        dwStyle = WS_POPUP;                                                                             //Windows Style
        ShowCursor(false);                                                                              //Panaikinti peles pointeri
        
        //Ijungti fullscreen
        //Ijungiama po duomenu uzkrovimo, nes ijungimas panaikina Win start juosta
        DEVMODE dmScreenSettings;                                                               //Device Mode
        memset(&dmScreenSettings, 0, sizeof(dmScreenSettings)); //Isvalyti atminti
        dmScreenSettings.dmSize = sizeof(dmScreenSettings);
        dmScreenSettings.dmPelsWidth    = langasX;                              //Lango plotis
        dmScreenSettings.dmPelsHeight   = langasY;                              //Lango aukstis
        dmScreenSettings.dmBitsPerPel   = langasBpp;                    //Bits per pixel
        dmScreenSettings.dmFields = DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
        if (ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL){
                #ifdef DEBUG
                Klaida("Parinkti vaizdo nustatymai nesuderinami su Jusu sistema");
                #endif
        return false;
        }

        AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);     //Nustatyti lango dydi

        //Kurti langa
        if (!(hWnd = CreateWindowEx(dwExStyle,                                                  //Extended Style For The Window
                                                                "OpenGL",                                                       //Class Name
                                                                langasPav,                                                      //Lango pavadinimas
                                                                dwStyle |                                                       //Defined Window Style
                                                                WS_CLIPSIBLINGS |                                       //Required Window Style
                                                                WS_CLIPCHILDREN,                                        //Required Window Style
                                                                0, 0,                                                           //Lango pozicija
                                                                langasX,                                                        //Lango ilgis
                                                                langasY,                                                        //Lango aukstis
                                                                NULL,                                                           //No Parent Window
                                                                NULL,                                                           //No Menu
                                                                hInstance,                                                      //Instance
                                                                NULL)))                                                         //Dont Pass Anything To WM_CREATE
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida kuriant langa");
                #endif
                return false;
        }

        static  PIXELFORMATDESCRIPTOR pfd =                             //pfd Tells Windows How We Want Things To Be
        {
                sizeof(PIXELFORMATDESCRIPTOR),                          //Size Of This Pixel Format Descriptor
                1,                                                                                      //Version Number
                PFD_DRAW_TO_WINDOW |                                            //Format Must Support Window
                PFD_SUPPORT_OPENGL |                                            //Format Must Support OpenGL
                PFD_DOUBLEBUFFER,                                                       //Must Support Double Buffering
                PFD_TYPE_RGBA,                                                          //Request An RGBA Format
                langasBpp,                                                                      //Bits per pixel
                0, 0, 0, 0, 0, 0,                                                       //Color Bits Ignored
                0,                                                                                      //No Alpha Buffer
                0,                                                                                      //Shift Bit Ignored
                32,                                                                                     //Accumulation Buffer
                0, 0, 0, 0,                                                                     //Accumulation Bits Ignored
                langasDepth,                                                            //Z-Buferio gylis 
                8,                                                                                      //Stencil Buffer
                0,                                                                                      //No Auxiliary Buffer
                PFD_MAIN_PLANE,                                                         //Main Drawing Layer
                0,                                                                                      //Reserved
                0, 0, 0                                                                         //Layer Masks Ignored
        };
        
        if (!(hDC = GetDC(hWnd)))                                               //Ar gautas Device Context?
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida nustatant 'GL Device Context'");
                #endif
                return false;
        }

        if (!(PixelFormat = ChoosePixelFormat(hDC, &pfd)))      //Ar gautas Pixel Format?
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida pasirenkant PixelFormat");
                #endif
                return false;
        }

        if(!SetPixelFormat(hDC, PixelFormat, &pfd))             //Ar pavyko nustatyti Pixel Format?
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida nustatant PixelFormat");
                #endif
                return false;
        }

        if (!(hRC = wglCreateContext(hDC)))                             //Ar gautas Rendering Context?
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida kuriant 'GL Rendering Context'");
                #endif
                return false;
        }

        if(!wglMakeCurrent(hDC, hRC))                                   //Ar aktyvuotas Rendering Context
        {
                LangasNaikinti();
                #ifdef DEBUG
                Klaida("Klaida aktyvuojant 'GL Rendering Context'");
                #endif
                return false;
        }
        
        
        //Inicializuoti OpenGL parametrus
        glShadeModel(GL_SMOOTH);
        glClearColor(0.0, 0.0, 0.0, 1.0);
        glClearDepth(1.0);
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LEQUAL);
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        //glEnable(GL_CULL_FACE);
        //glCullFace(GL_BACK);
        glEnable(GL_NORMALIZE);
        //Nustatyti ruka
        glEnable(GL_FOG);
        glFogi(GL_FOG_MODE, GL_EXP2);
        glFogf(GL_FOG_DENSITY, rukasKiek);
        glFogfv(GL_FOG_COLOR, rukasSp);
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
        glHint(GL_FOG_HINT, GL_NICEST);
        glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
        glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
        //Nustatyti apsvietima
        glLightfv(GL_LIGHT0, GL_POSITION, apsvPoz);
        glLightfv(GL_LIGHT0, GL_AMBIENT, apsvAmb);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, apsvDif);
        glLightfv(GL_LIGHT0, GL_SPECULAR, apsvSpec);
        //Sukurti viewport'a
        glViewport(0, 0, langasX, langasY);
        //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

        return true;
}

LRESULT CALLBACK WndProc(       HWND    hWnd,                   //Handle For This Window
                                                        UINT    uMsg,                   //Message For This Window
                                                        WPARAM  wParam,                 //Additional Message Information
                                                        LPARAM  lParam)                 //Additional Message Information
{
        switch (uMsg){
                case WM_ACTIVATE:                                                       //Watch For Window Activate Message
                        if (!HIWORD(wParam))                                    //Check Minimization State
                                langasBaigti = false;
                        else
                                langasBaigti = true;
                        return 0;

                case WM_SYSCOMMAND:                                                     //Intercept System Commands
                        switch (wParam)                                                 //Check System Calls
                        {
                                case SC_SCREENSAVE:                                     //Jungiamas Sreensaveris?
                                case SC_MONITORPOWER:                           //Monitorius uzmigo?
                                langasBaigti = true;
                                return 0;
                        }
                        break;

                case WM_CLOSE:                                                          //Gauta zinute Close?
                case WM_QUIT:                                                           //Gauta zinute Quit
                        langasBaigti = true;
                        return 0;

                case WM_KEYDOWN:                                                        //Paspaustas spaudukas?
                        spaud[wParam] = true;
                        Paspaudimai(wParam);
                        return 0;

                case WM_KEYUP:                                                          //Atleistas spaudukas?
                        spaud[wParam] = false;
                        return 0;
                
                #ifdef WAV
                case MM_WOM_OPEN:
                        mmRez = waveOutPrepareHeader(wavOut, &wavHeader, sizeof(wavHeader));
                        if(mmRez != MMSYSERR_NOERROR)
                                langasBaigti = true;
                        //mmRez = waveOutWrite(wavOut, &wavHeader, sizeof(wavHeader));
                        //if(mmRez != MMSYSERR_NOERROR)
                        //      langasBaigti = true;
                        return 0;
                #endif
        }

        return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

int WINAPI WinMain(     HINSTANCE       hInstance,
                                        HINSTANCE       hPrevInstance,
                                        LPSTR           lpCmdLine,
                                        int                     nCmdShow)
{
        MSG             msg;
        srand(time(NULL));
        laikas.Pradeti();

        if(!LangasKurti()){
                return 0;
        }
        
        //Rodyti langa
        ShowWindow(hWnd, SW_SHOW);
        SetForegroundWindow(hWnd);
        SetFocus(hWnd);
        
        if(!Uzkrauti()){
                return 0;
        }
        
        mPlay(&midiData);

        while(!langasBaigti){
                //Pirmiausia apdoroti visas gautas zinutes
                if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)){
                        TranslateMessage(&msg);
                        DispatchMessage(&msg);
                }
                //Naujas kadras
                else{
                        laikas.Gauti();
                        Valdymas();
                        Atnaujinti();
                        Piesti();
                        SwapBuffers(hDC);
                }
        }

        //Baigti
        mStop();
        //Naikinti();
        //LangasNaikinti();
        return (msg.wParam);
}
